# The task of the classes and functions here is to convert
# JSON encoded instrument setups into an
from __future__ import print_function, unicode_literals, absolute_import, division
from collections import OrderedDict
from itertools import islice

from ..gtc.headers import create_header_from_telpars


def get_obsmode(setup_data):
    mode = setup_data['appdata']['app']
    if mode == 'FullFrame':
        return FullFrame(setup_data)
    elif mode == 'Windows':
        return Windows(setup_data)
    elif mode == 'Drift':
        return Drift(setup_data)
    elif mode == 'Idle':
        return Idle()
    else:
        raise ValueError('Unrecognised mode: {}'.format(mode))


class ObsMode(object):

    def __init__(self, setup_data):
        """
        The base class for all HiPERCAM modes.

        Parameters
        ----------
        setup_data : dict
            Dictionary of HiPerCAM setup data
        """
        app_data = setup_data['appdata']
        nu, ng, nr, ni, nz = app_data['multipliers']
        dummy = app_data.get('dummy_out', 0)  # works even if dummy not set in app, default 0
        fastclk = app_data.get('fast_clks', 0)
        oscany = app_data.get('oscany', 0)

        # look in dictionary to see if we should use GPS hardware
        # default is to assume GPS attached
        self.use_gps_hardware = setup_data.get('gps_attached', 1)

        clockfile = 'hipercam.bclk'
        if not dummy:
            clockfile = 'hipercam_se.bclk'
        elif fastclk:
            clockfile = 'hipercam_fastclk.bclk'

        nodpattern = app_data.get('nodpattern', {})
        self.finite = app_data['numexp']
        self.detpars = {
            'DET.RESTRICT': 'F',
            'DET.SPEED': 0 if app_data['readout'] == 'Slow' else 1,
            'DET.BINX1': app_data['xbin'],
            'DET.BINY1': app_data['ybin'],
            'DET.CLRCCD': 'T' if app_data['clear'] else 'F',
            'DET.NCLRS': 5,
            'DET.DUMMY': dummy,
            'DET.FASTCLK': fastclk,
            'DET.EXPLED': 'T' if app_data['led_flsh'] else 'F',
            'DET.GPS': 'T',
            'DET.INCPRSCX': 'T' if app_data['oscan'] else 'F',
            'DET.INCOVSCY': 'T' if oscany else 'F',
            'DET.NSKIPS1': nu-1,
            'DET.NSKIPS2': ng-1,
            'DET.NSKIPS3': nr-1,
            'DET.NSKIPS4': ni-1,
            'DET.NSKIPS5': nz-1,
            'DET.SEQ.CLKFILE': clockfile,
            'DET.SEQ1.DIT': app_data['exptime'],
            'DET.SEQ.TRIGGER': 'T' if nodpattern else 'F',  # wait for trigger if nodding
            'DET.TDELAY.GUI': 1000*app_data['dwell']
        }

        # parameters for user-defined headers
        user_data = setup_data.get('user', {})
        imagetyp = user_data.get('flags', '')

        # put OBSTYPE in for GTC. Also, correctly set object
        obstype = 'Imaging'
        obj = user_data.get('target', '')
        if imagetyp == 'bias':
            obj = 'Bias'
            obstype = 'Bias'
        elif imagetyp == 'dark':
            obj = 'Dark'
            obstype = 'Dark'
        elif imagetyp == 'flat':
            obj = 'Skyflat'
            obstype = 'SkyFlat'

        userpars = []

        # gtc headers
        telpars = setup_data.get('gtc_headers', {})
        gtc_header_info = create_header_from_telpars(telpars)
        if gtc_header_info:
            userpars.extend(
                [(item, gtc_header_info[item]) for item in gtc_header_info]
            )
            gtcprgid = user_data.get('ID', '')

            gtcobid = user_data.get('OB', '1')
            # force 4 digit formatting
            try:
                gtcobid = '{:04d}'.format(int(gtcobid))
            except ValueError:
                # not an integer
                pass

            # force for calibrations
            if obj in ('Skyflat', 'Bias', 'Dark'):
                gtcprgid = 'Calib'

            # now force GTCOBID to be 'CALIB' if PRGID is
            # p.s this is a request from GTC staff
            if gtcprgid.lower().startswith('calib'):
                gtcobid = 'CALIB'
            # add prgid and obid to headers
            userpars.extend([
                ('GTCPRGID', gtcprgid),
                ('GTCOBID', gtcobid)
            ])

        # now update with GUI values. Do this after to allow override of
        # GTC telescope server headers
        userpars.extend([
            ('OBSERVER', user_data.get('Observers', '')),
            ('OBJECT', obj),
            ('RUNCOM', user_data.get('comment', '')),
            ('IMAGETYP', imagetyp),
            ('OBSMODE', obstype),
            ('FILTERS', user_data.get('filters', '')),
            ('PROGRM', user_data.get('ID', '')),
            ('PI', user_data.get('PI', ''))
        ])

        # data from TCS
        tcs_data = setup_data.get('tcs', {})
        userpars.extend([
            ('TELESCOP', tcs_data.get('tel', 'WHT')),
            ('RA', tcs_data.get('RA', '00:00:00.00')),
            ('DEC', tcs_data.get('DEC', '+00:00:00.0')),
            ('ELEVAT', tcs_data.get('alt', -999)),
            ('AZIMUTH', tcs_data.get('az', -999)),
            ('AIRMASS', tcs_data.get('secz', -999)),
            ('INSTRPA', tcs_data.get('pa', -999)),
            ('TELFOCUS', tcs_data.get('foc', -999)),
            ('MOONDIST', tcs_data.get('mdist', -999))
        ])
        if tcs_data.get('tel', 'WHT') == 'GTC':
            userpars.extend([('ORIGIN', 'GRANTECAN')])
        else:
            userpars.extend([('ORIGIN', 'ING')])

        # data from h/w monitoring processes
        hw_data = setup_data.get('hardware', {})
        userpars.extend([
            ('CCD1TEMP', hw_data.get('ccd1temp', -999)),
            ('CCD2TEMP', hw_data.get('ccd2temp', -999)),
            ('CCD3TEMP', hw_data.get('ccd3temp', -999)),
            ('CCD4TEMP', hw_data.get('ccd4temp', -999)),
            ('CCD5TEMP', hw_data.get('ccd5temp', -999)),
            ('CCD1VAC', hw_data.get('ccd1vac', -999)),
            ('CCD2VAC', hw_data.get('ccd2vac', -999)),
            ('CCD3VAC', hw_data.get('ccd3vac', -999)),
            ('CCD4VAC', hw_data.get('ccd4vac', -999)),
            ('CCD5VAC', hw_data.get('ccd5vac', -999)),
            ('CCD1FLOW', hw_data.get('ccd1flow', -999)),
            ('CCD2FLOW', hw_data.get('ccd2flow', -999)),
            ('CCD3FLOW', hw_data.get('ccd3flow', -999)),
            ('CCD4FLOW', hw_data.get('ccd4flow', -999)),
            ('CCD5FLOW', hw_data.get('ccd5flow', -999)),
            ('FPSLIDE', hw_data.get('fpslide', -999))
        ])
        # getting pars from GTC header info and TCS/GUI can insert duplicates
        # use a dict to remove duplicates
        self.userpars = OrderedDict(userpars)

    def setup_acq_task(self, nq=5):
        gps = 1 if self.use_gps_hardware else 0
        self.acq_dict = dict(nq=nq, gps=gps)
        if self.readoutMode == 4:
            # drift mode
            self.acq_dict['drf'] = 1
        else:
            self.acq_dict['drf'] = 0

        if self.detpars['DET.SPEED'] == 0:
            # slow
            self.acq_dict['aveg'] = 4
        else:
            # fast
            self.acq_dict['aveg'] = 1

    @property
    def acq_command(self):
        template = 'acqproc hiperCamCCD -chip 5 -t 4 -nq {nq} -deml 0 -gps {gps} -aveg {aveg} -aps 8 -drf {drf}'
        return template.format(**self.acq_dict)

    @property
    def readmode_command(self):
        return 'setup DET.READ.CURID {}'.format(self.readoutMode)

    @property
    def setup_command(self):
        setup_string = 'setup'

        for key in self.detpars:
            setup_string += ' {} {} '.format(key, self.detpars[key])

        if self.finite:
            setup_string += ' DET.FRAM2.BREAK {} '.format(self.finite)

        return setup_string

    @property
    def header_commands(self):
        def chunks(data, SIZE=10):
            it = iter(data)
            for i in range(0, len(data), SIZE):
                yield {k: data[k] for k in islice(it, SIZE)}

        setup_strings = []
        for chunk in chunks(self.userpars):
            setup_string = 'setup'
            for key in chunk:
                if chunk[key] != '':
                    value = chunk[key]
                    # add quotes to strings with spaces
                    try:
                        if ' ' in value:
                            value = '"' + value + '"'
                    except:
                        pass
                    setup_string += ' {} {} '.format(key, value)
            setup_strings.append(setup_string)
        return setup_strings


class FullFrame(ObsMode):
    def __init__(self, setup_data):
        super(FullFrame, self).__init__(setup_data)
        self.readoutMode = 1
        self.setup_acq_task(nq=5)


class Idle(ObsMode):
    """
    An idle mode to keep clearing the chip and stop taking GPS timestamps.
    """
    def __init__(self):
        app_data = {
            'xbin': 1,
            'ybin': 1,
            'clear': True,
            'readout': 'Slow',
            'led_flsh': False,
            'oscan': False,
            'dummy': 1,
            'numexp': 0,
            'oscany': False,
            'multipliers': (1, 1, 1, 1, 1),
            'exptime': 10,
            'dwell': 10
        }
        setup_data = {'appdata': app_data}
        super(Idle, self).__init__(setup_data)
        self.detpars['DET.GPS'] = 'F'
        self.readoutMode = 1  # FF, Slow
        self.setup_acq_task(nq=5)


class Windows(ObsMode):
    def __init__(self, setup_data):
        super(Windows, self).__init__(setup_data)
        app_data = setup_data['appdata']
        win1 = {
            'DET.WIN1.NX': app_data['x1size'],
            'DET.WIN1.NY': app_data['y1size'],
            'DET.WIN1.XSE': app_data['x1start_lowerleft'] - 1,
            'DET.WIN1.XSF': 2049 - app_data['x1start_lowerright'] - app_data['x1size'],
            'DET.WIN1.XSG': 2049 - app_data['x1start_upperright'] - app_data['x1size'],
            'DET.WIN1.XSH': app_data['x1start_upperleft'] - 1,
            'DET.WIN1.YS': app_data['y1start'] - 1,
            'DET.WIN1.YS.GUI': app_data['y1start'],
            'DET.WIN1.XSLL': app_data['x1start_lowerleft'],
            'DET.WIN1.XSLR': app_data['x1start_lowerright'],
            'DET.WIN1.XSUL': app_data['x1start_upperleft'],
            'DET.WIN1.XSUR': app_data['x1start_upperright']
        }
        self.detpars.update(win1)

        if 'x2size' in app_data:
            win2 = {
                'DET.WIN2.NX': app_data['x2size'],
                'DET.WIN2.NY': app_data['y2size'],
                'DET.WIN2.XSE': app_data['x2start_lowerleft'] - 1,
                'DET.WIN2.XSF': 2049 - app_data['x2start_lowerright'] - app_data['x2size'],
                'DET.WIN2.XSG': 2049 - app_data['x2start_upperright'] - app_data['x2size'],
                'DET.WIN2.XSH': app_data['x2start_upperleft'] - 1,
                'DET.WIN2.YS': app_data['y2start'] - 1,
                'DET.WIN2.YS.GUI': app_data['y2start'],
                'DET.WIN2.XSLL': app_data['x2start_lowerleft'],
                'DET.WIN2.XSLR': app_data['x2start_lowerright'],
                'DET.WIN2.XSUL': app_data['x2start_upperleft'],
                'DET.WIN2.XSUR': app_data['x2start_upperright']
            }
            self.detpars.update(win2)

        if self.readoutMode == 2:
            self.setup_acq_task(nq=30)
        else:
            self.setup_acq_task(nq=10)

    @property
    def readoutMode(self):
        if 'DET.WIN2.NX' in self.detpars:
            mode = 3
        else:
            mode = 2
        return mode


class Drift(ObsMode):
    def __init__(self, setup_data):
        super(Drift, self).__init__(setup_data)
        app_data = setup_data['appdata']
        win1 = {
            'DET.DRWIN.NX': app_data['x1size'],
            'DET.DRWIN.NY': app_data['y1size'],
            'DET.DRWIN.YS': app_data['y1start'] - 1,
            'DET.DRWIN.YS.GUI': app_data['y1start'],
            'DET.DRWIN.XSE': app_data['x1start_left'] - 1,
            'DET.DRWIN.XSF': 2049 - app_data['x1start_right'] - app_data['x1size'],
            'DET.DRWIN.XSH': app_data['x1start_left'] - 1,
            'DET.DRWIN.XSG': 2049 - app_data['x1start_right'] - app_data['x1size'],
            'DET.DRWIN.XSL': app_data['x1start_left'],
            'DET.DRWIN.XSR': app_data['x1start_right'],
        }
        self.detpars.update(win1)
        self.nrows = 520  # number of rows in storage area
        self.readoutMode = 4
        self.setup_acq_task(nq=800)

    @property
    def num_stacked(self):
        """
        Number of windows stacked in frame transfer area
        """
        ny = self.detpars['DET.DRWIN.NY']
        return int((self.nrows/ny + 1) / 2)

    @property
    def pipe_shift(self):
        """
        Extra shift to add to some windows to ensure uniform exposure time.

        Returned in units of vertical clocks. Should be multiplied by the
        vclock time to obtain pipe_shift in seconds.
        """
        ny = self.detpars['DET.DRWIN.NY']
        return (self.nrows - (2*self.num_stacked - 1)*ny)



# COMPO components
import astropy.units as u
from astropy.time import Time
from twisted.logger import Logger

from ..devices.properties import microstep
from ..devices.newport import NewportESP301
from ..machines import create_machine

equivalencies = [
    # from, to, forward, backward (assume 1-1 for now)
    (u.rad, microstep, lambda x: 1.0, lambda x: 1.0),
    (u.mm, microstep, lambda x: 1.0, lambda x: 1.0)
]

# match order in NewportESP301
esp_axes = {1: 'pickoff', 2: 'injection', 3: 'lens'}


class Compo(object):
    """
    Model for COMPO hardware.

    COMPO is controlled by a Newport ESP301 controller, which has 3 axes.

    These axes are the angles of the pickoff arm and injection arm and the
    linear lens slide.
    """
    log = Logger()

    # can be set by publishing to a device level topic of same name
    settable_attributes = ('target_injection_angle',
                           'target_pickoff_angle',
                           'target_lens_position')

    # all state change triggers are automatically addded as RPC calls
    # add any extra methods you want exposed here
    extra_rpc_calls = ('power_on_axis', 'power_off_axis')

    def __init__(self, port, emulate=False):

        self.dev = NewportESP301(port, emulate=emulate, unit_equivalencies=equivalencies)

        # state machines to track state. Controller for connections
        self._controller = create_machine('connections',
                                          initial_context={'device': self})
        self.machines = dict(connection=self._controller)
        self._controller.clock.start()

        # axes!
        # ORDER HERE MUST MATCH ORDER IN Newport!
        for ax in esp_axes:
            name = esp_axes[ax]
            ctx = {
                'device': self.dev.axis[ax-1],
                'current': 0*u.mm if name == 'lens' else 0*u.deg,
                'target': 0*u.mm if name == 'lens' else 0*u.deg,
                'poll_time': 2
            }
            axis = create_machine('motors', initial_context=ctx)
            axis.clock.start()
            self.machines[name] = axis
            # bind axes to receive connect disconnect from controller
            self._controller.bind(axis)

    @property
    def current_injection_angle(self):
        return self.machines['injection'].context['current']

    @property
    def target_injection_angle(self):
        return self.machines['injection'].context['target']

    @target_injection_angle.setter
    def target_injection_angle(self, value):
        print('setting injection angle')
        if not isinstance(value, u.Quantity):
            value = value * u.deg
        self.machines['injection'].queue('positionSet', position=value)

    @property
    def current_pickoff_angle(self):
        return self.machines['pickoff'].context['current']

    @property
    def target_pickoff_angle(self):
        return self.machines['pickoff'].context['target']

    @target_pickoff_angle.setter
    def target_pickoff_angle(self, value):
        if not isinstance(value, u.Quantity):
            value = value * u.deg
        self.machines['pickoff'].queue('positionSet', position=value)

    @property
    def current_lens_position(self):
        return self.machines['lens'].context['current']

    @property
    def target_lens_position(self):
        return self.machines['lens'].context['target']

    @target_lens_position.setter
    def target_lens_position(self, value):
        if not isinstance(value, u.Quantity):
            value = value * u.mm
        self.machines['lens'].queue('positionSet', position=value)

    def telemetry(self):
        """
        Called periodically to provide a telemetry package for the device
        """
        ts = Time.now()
        tel = dict(timestamp=ts)
        state = {key: self.machines[key].configuration for key in self.machines}
        tel['state'] = state
        tel['pickoff_angle'] = dict(current=self.current_pickoff_angle,
                                    target=self.target_pickoff_angle)
        tel['injection_angle'] = dict(current=self.current_injection_angle,
                                      target=self.target_injection_angle)
        tel['lens_position'] = dict(current=self.current_lens_position,
                                    target=self.target_lens_position)
        return tel

    # Methods REQUIRED for controller state machine
    def try_connect(self):
        try:
            self.dev.connect()
            # update device in machines
            for ax in esp_axes:
                name = esp_axes[ax]
                machine = self.machines[name]
                axis = self.dev.axis[ax-1]
                machine.context['device'] = axis

        except Exception:
            self.error()
            return False
        return True

    def disconnect(self):
        self.dev.disconnect()

    # additional RPC calls
    def power_on_axis(self, axis):
        self.dev.axis[axis].power_on()

    def power_off_axis(self, axis):
        self.dev.axis[axis].power_off()



import pkg_resources
import os

from sismic.interpreter import Interpreter
from sismic.io import import_from_yaml


def get_yaml_file(machine_name):
    if not machine_name.endswith(".yaml"):
        machine_name += ".yaml"
    path = os.path.join('hcam_devices/machines', machine_name)
    full_path = pkg_resources.resource_filename(__name__, path)
    if not os.path.exists(full_path):
        full_path = os.path.join(os.path.dirname(__file__), machine_name)
    return full_path


def create_machine(name, **kwargs):
    full_path = get_yaml_file(name)
    return Interpreter(import_from_yaml(filepath=full_path), **kwargs)



# focal plane slide
import astropy.units as u
from astropy.time import Time
from twisted.logger import Logger

from ..devices.properties import microstep
from ..devices.zaber import ZaberLS, ZaberLSEmulator
from ..machines import create_machine

# the next define ranges for the movement in terms of
# microsteps, millimetres and pixels
MIN_MS = 0
MAX_MS = 1664000
MIN_PX = 2760.2
MAX_PX = -1155.3
MM_PER_MS = 0.000078
MIN_MM = MM_PER_MS*MIN_MS
MAX_MM = MM_PER_MS*MAX_MS

# Standard pixel positions for unblocking and blocking the CCD
UNBLOCK_POS = 1100.
BLOCK_POS = -100.


def mm_to_ms(value):
    return MIN_MS + int((MAX_MS - MIN_MS) * (value - MIN_MM) / (MAX_MM-MIN_MM) + 0.5)


def ms_to_mm(value):
    return MIN_MM + (MAX_MM - MIN_MM)*(value - MIN_MS)/(MAX_MS - MIN_MS)


def px_to_ms(value):
    return MIN_MS + int((MAX_MS-MIN_MS) * (value - MIN_PX) / (MAX_PX - MIN_PX) + 0.5)


def ms_to_px(value):
    return MIN_PX + (MAX_PX - MIN_PX)*(value - MIN_MS)/(MAX_MS - MIN_MS)


unit_scale = [
    # from, to, forward, backward
    (u.pix, microstep, px_to_ms, ms_to_px),
    (u.mm, microstep, mm_to_ms, ms_to_mm)
]


class FocalPlaneSlide(object):
    """
    Focal plane slide.

    Uses a Motor state machine to track state, and a Zaber linear stage device to send
    messages.

    Periodically updates WAMP server with updates of position and state.

    All positions in pixels
    """

    log = Logger()

    # can be set by publishing to device level topic of same name
    settable_attributes = ('target_position',)
    # all state change triggers are automatically addded as RPC calls
    # add any extra methods you want exposed here
    extra_rpc_calls = ('block', 'unblock', 'enable', 'disable', 'reset')

    def __init__(self, ip_address, port, emulate=True):
        if emulate:
            self.dev = ZaberLSEmulator(ip_address, port, unit_equivalencies=unit_scale)
        else:
            self.dev = ZaberLS(ip_address, port, unit_equivalencies=unit_scale)

        # state machines to keep track of state
        self._controller = create_machine('connections',
                                          initial_context={'device': self})
        self._axis = create_machine('motors',
                                    initial_context={'device': self, 'current': 0*u.pix,
                                                     'target': 0*u.pix, 'poll_time': 1})
        self._controller.clock.start()
        self._axis.clock.start()

        self.machines = {'connection': self._controller, 'stage': self._axis}
        # axis will recieve events from controller
        self._controller.bind(self._axis)

    @property
    def current_position(self):
        return self._axis.context['current']

    @property
    def target_position(self):
        return self._axis.context['target']

    @target_position.setter
    def target_position(self, value):
        if not isinstance(value, u.Quantity):
            value = value * u.pix
        self._axis.queue('positionSet', position=value)

    def block(self):
        self.target_position = BLOCK_POS
        self._axis.queue('move')

    def unblock(self):
        self.target_position = UNBLOCK_POS
        self._axis.queue('move')

    def enable(self):
        self.dev.enable()

    def disable(self):
        self.dev.disable()

    def reset(self):
        self.dev.reset()
        self._controller.queue('disconnect')

    def telemetry(self):
        """
        Called periodically to provide a telemetry package for the device
        """
        ts = Time.now()
        return dict(
            timestamp=ts,
            position=dict(current=self.current_position,
                          target=self.target_position),
            state={key: self.machines[key].configuration for key in self.machines}
        )

    # methods required for motor state machine
    # Keep these on slide object rather than moving to Zaber since
    # querying target position of Zaber stage is not implemented
    def home(self):
        try:
            self.dev.home()
        except Exception as err:
            self.log.error(str(err))

    def on_target(self):
        current = self.current_position.to(microstep, equivalencies=unit_scale)
        target = self.target_position.to(microstep, equivalencies=unit_scale)
        delta = abs((current-target).value)
        return int(delta) < 5

    def move(self, value):
        if not isinstance(value, u.Quantity):
            # assume pixels
            value = value * u.pix
        try:
            self.dev.move(value)
        except Exception as err:
            self.log.error(str(err))

    def stop(self):
        self.dev.stop()

    def try_connect(self):
        """
        Always needs to be defined.  Returns True on successful connection attempt.
        """
        # attempt connection, return true to allow device state to change
        try:
            self.dev.connect()
        except Exception:
            # send msg to broker
            self.error()
            return False
        return True

    def disconnect(self):
        self.dev.close()

    def poll(self):
        """
        State machines mean that this is automatically polled.

        Update current position and return True/False for moving state
        """
        current_position = self.dev.position.to(u.pix, equivalencies=unit_scale)
        return current_position, self.dev.moving



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import glob
import os
import sys

with open('README.rst', encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst', encoding='utf-8') as history_file:
    history = history_file.read()

requirements = [
    'astropy', 'requests', 'sismic', 'autobahn', 'pyserial',
    'twisted', 'influxdb', 'txaio', 'watchdog', 'six'
]
if sys.version_info[0] == 2:
    requirements.append('pymodbus')
else:
    requirements.append('pymodbus3')

test_requirements = [
    # TODO: put package test requirements here
]

# Treat everything in scripts except README.rst as a script to be installed
scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))
           if os.path.basename(fname) != 'README.rst']

setup(
    name='hcam_devices',
    version='0.2.1',
    description="Device Communication via WAMP for HiPerCAM",
    long_description=readme + '\n\n' + history,
    author="Stuart Littlefair",
    author_email='s.littlefair@shef.ac.uk',
    url='https://github.com/HiPERCAM/hcam_devices',
    download_url='https://github.com/HiPERCAM/hcam_devices/archive/v0.2.1.tar.gz',
    packages=[
        'hcam_devices',
        'hcam_devices.gtc',
        'hcam_devices.devices',
        'hcam_devices.components',
        'hcam_devices.machines',
        'hcam_devices.models',
        'hcam_devices.testing',
        'hcam_devices.wamp',
        'hcam_devices.wamp.utils',
        'hcam_devices.utils'
    ],
    package_dir={'hcam_devices':
                 'hcam_devices'},
    include_package_data=True,
    scripts=scripts,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='hcam_devices',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



# GTC
import time
from itertools import cycle
from astropy.time import Time
from astropy.io.fits import Header
from twisted.logger import Logger

from ..machines import create_machine
from ..utils.filesystem import FITSWatcher
from ..wamp.utils import call
from ..gtc.headers import create_header_from_telpars
try:
    from ..gtc.corba import TelescopeServer
except Exception:
    TelescopeServer = None


class FakeTelescopeServer(object):
    def __init__(self):
        self._inpos = True
        self._last_move = time.time()
        self.ra = 100
        self.dec = 20

    @property
    def ready(self):
        return True

    def in_position(self):
        if time.time() - self._last_move > 5:
            self._inpos = True
            self.ra += self.raoff
            self.raoff = 0
            self.dec += self.decoff
            self.decoff = 0
            return True

        return False

    def requestTelescopeOffset(self, dra, ddec):
        print('offsetting ', dra, ddec)
        self._last_move = time.time()
        self.raoff = dra
        self.decoff = ddec

    def getHumidity(self):
        return 0.2

    def getCabinetTemperature1(self):
        return 25.0

    def getTelescopeParams(self):
        hdr = Header()
        hdr['RADEG'] = (self.ra, 'RA (degrees)')
        hdr['DECDEG'] = (self.dec, 'DEC (degrees)')
        hdr['INSTRPA'] = (22.1, 'Rotator PA')
        hdr['M2UZ'] = (-2.13, 'Focus offset')
        return hdr.tostring(sep=';').split(';')[:-1]


class GTC(object):
    """
    Telescope communication with GTC
    """
    log = Logger()

    settable_attributes = ()

    # all state change triggers are automatically addded as RPC calls
    # in this case this is only "do_offset"
    # add any extra methods you want exposed here
    extra_rpc_calls = ('get_telescope_pars', 'get_rack_temperature', 'get_humidity',
                       'load_nod_pattern', 'track_new_run')

    # these calls will be performed repeatedly
    extra_looping_calls = {'poll_runs': 0.1}

    def __init__(self, emulate=True, path='/data'):
        print('emulation mode: ', emulate)
        if emulate:
            self._server = FakeTelescopeServer()
        else:
            self._server = TelescopeServer()
        self._gtcmachine = create_machine('gtc', initial_context={'server': self._server})
        # route events sent by machine to our event handler
        self._gtcmachine.bind(self.machine_event_handler)
        # start clock
        self._gtcmachine.clock.start()
        self.machines = {"gtc": self._gtcmachine}

        # define dummy nod patterns
        self.ra_offsets = cycle([0.0])
        self.dec_offsets = cycle([0.0])

        # setup polling of latest run
        self.path = path
        self.track_new_run()

    def poll_runs(self):
        self.run_watcher.poll()

    def machine_event_handler(self, event):
        if event.name == 'trigger_exposure':
            call('hipercam.ngc.rpc.trigger')

    def telemetry(self):
        """
        Called periodically to provide a telemetry package
        """
        ts = Time.now()
        try:
            telpars = create_header_from_telpars(self._gtcmachine.context['telpars'])
        except TypeError:
            telpars = {}
        return dict(
            timestamp=ts,
            telpars=telpars,
            rack_temp=self._gtcmachine.context['rack_temp'],
            humidity=self._gtcmachine.context['humidity'],
            ra_offset=self.cumulative_ra_offset,
            dec_offset=self.cumulative_dec_offset,
            state={key: self.machines[key].configuration for key in self.machines}
        )

    def get_relative_offsets(self, abs_ra_offset, abs_dec_offset):
        # convert an absolute offset into one relative to current position
        return (abs_ra_offset - self.cumulative_ra_offset,
                abs_dec_offset - self.cumulative_dec_offset)

    def get_next_offsets(self):
        return self.get_relative_offsets(next(self.ra_offsets), next(self.dec_offsets))

    def on_run_modified_callback(self):
        raoff, decoff = self.get_next_offsets()
        self._gtcmachine.queue('frame_written', raoff=raoff, decoff=decoff)
        self.cumulative_ra_offset += raoff
        self.cumulative_dec_offset += decoff

    def track_new_run(self):
        self.cumulative_ra_offset = 0.0
        self.cumulative_dec_offset = 0.0
        # create a polling object to watch the data directory
        self.run_watcher = FITSWatcher(self.path, self.on_run_modified_callback)

    def load_nod_pattern(self, raoffs, decoffs):
        try:
            self.ra_offsets = cycle(list(raoffs))
            self.dec_offsets = cycle(list(decoffs))
        except ValueError:
            raise ValueError('could not create offset patterns from input')

    def get_humidity(self):
        return self._gtcmachine.context['humidity']

    def get_rack_temperature(self):
        return self._gtcmachine.context['rack_temp']

    def get_telescope_pars(self):
        return self._gtcmachine.context['telpars']
